#
# EDK BSP board generation for U-boot supporting Microblaze and PPC
#
# (C) Copyright 2007 Xilinx, Inc.
# Based on original code:
# (C) Copyright 2007 Michal Simek
#
# Michal SIMEK <monstr@monstr.eu>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307 USA
#
# ChangeLOG:
# 2.00.b Remove problem with storing file
# 2.00.a Testing interrupt connection between Microblaze & INTC
# 1.01.c add simple FDT support - testing version
# 1.01.b add Microblaze HW_VER support, FLAGS, Uart16550. Remove MSS drivers
#	dependence, Timer handle throught INTC, BARREL shifter
# 1.01.a add MSR support and mch_opb_ddr
# 1.01.f add FSL support
# 1.01.e add IIC support
# 1.00.d clear Czech comments
# 1.00.c synchronization FLASH_MEMORY

# Globals variable
set drvlist [list]
set cpunumber 0
set periphery_array ""

proc uboot_drc {os_handle} {
	puts "\#--------------------------------------"
	puts "\# uboot BSP DRC...!"
	puts "\#--------------------------------------"
}

proc create_standalone_namespace {} {
    namespace eval standalone_bsp {
        global env
        set edk_path $env(XILINX_EDK)
        source ${edk_path}/sw/lib/bsp/standalone_v1_00_a/data/standalone_v2_1_0.tcl
    }
}

proc generate {os_handle} {
	generate_fdt $os_handle
	generate_uboot $os_handle

	global edk_path
	global env
	set edk_path $env(XILINX_EDK)
	create_standalone_namespace

	set orig_mbsrcdir "${edk_path}/sw/lib/bsp/standalone_v1_00_a/src/microblaze"
	set orig_ppcsrcdir "${edk_path}/sw/lib/bsp/standalone_v1_00_a/src/ppc405"
	set orig_profilesrcdir "${edk_path}/sw/lib/bsp/standalone_v1_00_a/src/profile"
	set mbsrcdir "./src/microblaze"
	set ppcsrcdir "./src/ppc405"
	set profilesrcdir "./src/profile"

	file mkdir "./src"
	file copy ${orig_mbsrcdir} $mbsrcdir
	file copy ${orig_ppcsrcdir} $ppcsrcdir
	file copy ${orig_profilesrcdir} $profilesrcdir

	puts "--------------------------------"
	puts "Calling standalone::generate"
	standalone_bsp::generate $os_handle
	puts "--------------------------------"
}

# procedure post_generate
# This generates the drivers directory for uboot
# and runs the ltypes script

proc post_generate {lib_handle} {
	set pwd [pwd]
	set name [xget_value $lib_handle "PARAMETER" name]
	if {$name != ""} {
		set uboot_dir "uboot/board/xilinx/$name"
		exec bash -c "mkdir -p $uboot_dir"
		exec bash -c "cp ../../include/xparameters.h $uboot_dir/xparameters-edk.h"
		exec bash -c "cp ../../../system.mhs $uboot_dir"
		exec bash -c "cp ../../../system.mss $uboot_dir"
		exec bash -c "mv xparameters.h $uboot_dir"
		exec bash -c "mv config.mk $uboot_dir"
	}

	standalone_bsp::post_generate ${lib_handle}
}


proc generate_fdt {os_handle} {
	puts "\#--------------------------------------"
	puts "\# FDT BSP generate..."
	puts "\#--------------------------------------"

	set tree {}
	set bootargs [xget_sw_parameter_value $os_handle "bootargs"]
	set chosen {}
	lappend chosen [list bootargs string $bootargs]
	lappend tree [list model string "testing"]
	lappend tree [list chosen tree $chosen]

	set proc_handle [xget_libgen_proc_handle]
	set hwproc_handle [xget_handle $proc_handle "IPINST"]
	set proctype [xget_value $hwproc_handle "OPTION" "IPNAME"]
	switch $proctype {
		"microblaze" {
			set intc [get_handle_to_intc $proc_handle "Interrupt"]
			set tree [gen_microblaze $tree $hwproc_handle [default_parameters $hwproc_handle]]
			set test [bus_bridge $tree $hwproc_handle $intc "DPLB" ""]
			if { "$test" != "$tree"} {
				set tree $test
				lappend tree [list compatible string "ibm,plb"]
			}
			set test [bus_bridge $tree $hwproc_handle $intc "DOPB" ""]
			if { "$test" != "$tree"} {
				set tree $test
				lappend tree [list compatible string "ibm,opb"]
			}
		}
		"ppc405" -
		"ppc405_virtex4" {
			set intc [get_handle_to_intc $proc_handle "EICC405EXTINPUTIRQ"]
			set tree [gen_ppc405 $tree $hwproc_handle [default_parameters $hwproc_handle]]
			set tree [bus_bridge $tree $hwproc_handle $intc "DPLB" ""]
			lappend tree [list compatible string "ibm,plb"]

		}
		default {
			error "unsupported CPU"
		}
	}
	lappend tree [list \#size-cells hexint 1]
	lappend tree [list \#address-cells hexint 1]

	set tree [gen_memories $tree $hwproc_handle]

	set tree_file [open "fdt.dts" w]
	headerc $tree_file
	puts -nonewline $tree_file "/ "
	write_tree 0 $tree_file $tree
	puts $tree_file " ;"
	close $tree_file
}

proc generate_uboot {os_handle} {
	# ******************************************************************************
	# U-BOOT part
	# ******************************************************************************

	puts "\#--------------------------------------"
	puts "\# uboot BSP generate..."
	puts "\#--------------------------------------"

	# Open files and print GPL licence
	set config_file2 [open "config.mk" w]
	headerm $config_file2
	set config_file [open "xparameters.h" w]
	headerc $config_file

	# ******************************************************************************
	# print system clock
	set proc_handle [xget_libgen_proc_handle]
	set sys_clock [xget_sw_parameter_value $proc_handle "CORE_CLOCK_FREQ_HZ"]
	puts $config_file "/* System Clock Frequency */"
	puts $config_file "#define XILINX_CLOCK_FREQ\t$sys_clock\n"

	# Microblaze
	set hwproc_handle [xget_handle $proc_handle "IPINST"]
	set args [xget_hw_parameter_handle $hwproc_handle "*"]
	set proctype [xget_value $hwproc_handle "OPTION" "IPNAME"]
	switch $proctype {
		"microblaze" {
			# write only name of instance
			puts $config_file "/* Microblaze is [xget_hw_parameter_value $hwproc_handle "INSTANCE"] */"

			foreach arg $args {
				set arg_name [xget_value $arg "NAME"]
				set arg_name [string map -nocase {C_ ""} $arg_name]
				set arg_value [xget_value $arg "VALUE"]
				#	puts $config_file "DEBUG $arg_name $arg_value"
				switch $arg_name {
					USE_MSR_INSTR {
						puts $config_file "#define XILINX_USE_MSR_INSTR\t$arg_value"
					}
					FSL_LINKS {
						puts $config_file "#define XILINX_FSL_NUMBER\t$arg_value"
					}
					HW_VER {
						puts $config_file2 "PLATFORM_CPPFLAGS += -mcpu=v$arg_value"
					}
					USE_BARREL {
						if {[string match $arg_value "1"]} {
							puts $config_file2 "PLATFORM_CPPFLAGS += -mxl-barrel-shift"
						}
					}
					USE_DIV {
						if {[string match $arg_value "1"]} {
							puts $config_file2 "PLATFORM_CPPFLAGS += -mno-xl-soft-div"
						}
						#FIXME What is -mno-xl-hard-div
					}
					USE_HW_MUL {
						if {[string match $arg_value "1"]} {
							 puts $config_file2 "PLATFORM_CPPFLAGS += -mno-xl-soft-mul"
						} else {
							 puts $config_file2 "PLATFORM_CPPFLAGS += -mxl-soft-mul"
						}
						#FIXME What is -mxl-multiply-high???
					}
					USE_PCMP_INSTR {
						if {[string match $arg_value "1"]} {
							puts $config_file2 "PLATFORM_CPPFLAGS += -mxl-pattern-compare"
						}
					}
					USE_FPU {
						if {[string match $arg_value "1"]} {
							puts $config_file2 "PLATFORM_CPPFLAGS += -mhard-float"
						}
					}
					PVR {
						puts $config_file "#define XILINX_PVR\t$arg_value"
					}
					default {}
				}
			}

			# OPB bus resolve
			set dopb2 [xget_handle $hwproc_handle "BUS_INTERFACE" "DOPB"]
			set dopb [xget_value $dopb2 "VALUE"]
			set iopb [xget_handle $hwproc_handle "BUS_INTERFACE" "IOPB"]
			set iopb [xget_value $iopb "VALUE"]
			if { $dopb == $iopb } {
				set system_bus "$dopb"
				puts "System bus for instruction and data $dopb"
				#testing
				#	set bus [xget_sw_parameter_value $os_handle "opb_v20"]
				##	set bus_handle [xget_sw_ipinst_handle_from_processor $proc_handle $bus]
				##	set hodn [xget_sw_parameter_value $bus_handle "C_EXT_RESET_HIGH"]
				#	puts "fdf $bus fds"
				#	set clk [xget_handle $dopb "PORT" "OPB_Clk"]
				#	puts "$clk"
				#	set clk [xget_value $clk "VALUE"]
				#	error "$clk"
				#end testing
			} else {
				error "different microblaze architecture - dual busses $iopb $dopb"
			}
			puts $config_file ""
			uboot_intc $os_handle $proc_handle $config_file $config_file2 $system_bus
		}
		"ppc405" -
		"ppc405_virtex4" {
			puts "unsupported processor type $proctype\n"
			error "ERROR $proctype not supported by U-BOOT yet"
		}
		default {
			error "This type of CPU is not supported by U-BOOT yet"
		}

	}
	close $config_file
	close $config_file2
}

proc uboot_intc {os_handle proc_handle config_file config_file2 system_bus} {
	# ******************************************************************************
	# Interrupt controler
	set intc [xget_sw_parameter_value $os_handle "intc"]
	if {[string match "" $intc] || [string match -nocase "none" $intc]} {
		puts $config_file "/* Interrupt controller not defined */"
	} else {
		puts $config_file "/* Interrupt controller is $intc */"
		set intc_handle [xget_sw_ipinst_handle_from_processor $proc_handle $intc]
		#test for correct system bus
		test_buses $system_bus $intc_handle "SOPB"
		# Interrupt controller address
		set intc_base [xget_sw_parameter_value $intc_handle "C_BASEADDR"]
		set intc_end [xget_sw_parameter_value $intc_handle "C_HIGHADDR"]
		set intc_base [format "0x%08x" $intc_base]
		puts $config_file "#define XILINX_INTC_BASEADDR\t$intc_base"
		set intc_value [xget_sw_parameter_value $intc_handle "C_NUM_INTR_INPUTS"]

		puts $config_file "#define XILINX_INTC_NUM_INTR_INPUTS\t$intc_value"
		set intc_value [expr $intc_value - 1]
		set intc_handle [xget_sw_ipinst_handle_from_processor $proc_handle $intc]
		set port_list [xget_port_by_subtype $intc_handle "*"]
		foreach port $port_list {
			set name [xget_value $port "NAME"]
			if {[string match -nocase $name "intr"]} {
				set intc_irq [xget_value $port "VALUE"]
				#DEBUG	puts $config_file "/* pripojene interrupty $name=$intc_irq */"
				set intc_signals [split $intc_irq "&"]
				# spit the signals
				#DEBUG	puts $config_file "$signals"
			}
		}
		puts $config_file ""
	}
	# ****************************************************************************
	# Timer part
	# handle timer if exists intc
	set timer [xget_sw_parameter_value $os_handle "timer"]
	if {[string match "" $timer] || [string match -nocase "none" $timer]} {
		puts $config_file "/* Timer not defined */"
	} else {
		set timer_handle [xget_sw_ipinst_handle_from_processor $proc_handle $timer]
		#test for correct system bus
		test_buses $system_bus $timer_handle "SOPB"
		set timer_base [xget_sw_parameter_value $timer_handle "C_BASEADDR"]
		set timer_end [xget_sw_parameter_value $timer_handle "C_HIGHADDR"]
		set timer_base [format "0x%08x" $timer_base]
		puts $config_file "/* Timer pheriphery is $timer */"
		puts $config_file "#define XILINX_TIMER_BASEADDR\t$timer_base"

		#	puts "$timer_handle $intc $intc_value $intc_signals Interrupt"
		set intr [get_intr $timer_handle $intc_handle $intc_value "Interrupt"]
		puts $config_file "#define XILINX_TIMER_IRQ\t$intr"
	}
	puts $config_file ""
	# ******************************************************************************
	# UartLite driver - I suppose, only uartlite driver
	set uart [xget_sw_parameter_value $os_handle "stdin_out"]
	if {[string match "" $uart] || [string match -nocase "none" $uart]} {
		puts $config_file "/* Uart not defined */"
		puts "ERROR Uart not specified. Please specific console"
	} else {
		set uart_handle [xget_sw_ipinst_handle_from_processor $proc_handle $uart]
		puts $config_file "/* Uart pheriphery is $uart */"
		set type [xget_value $uart_handle "VALUE"]
		switch $type {
			"opb_uart16550" {
				puts $config_file "#define XILINX_UART16550"
				set uart_baseaddr [xget_sw_parameter_value $uart_handle "C_BASEADDR"]
				set uart_baseaddr [format "0x%08x" $uart_baseaddr]
				puts $config_file "#define XILINX_UART16550_BASEADDR\t$uart_baseaddr"
				# FIXME bad sysclock frequency
				#	puts $config_file "#define XILINX_UART16550_CLOCK_HZ\t$system_bus"
				#it is driver dependent code
				set drv_handle [xget_handle $proc_handle "DRIVER" "uartns550"]
				if { [llength $drv_handle] == 0 } {
					error "Please specify your standalone driver for Uart16550"
				}
				puts $config_file "#define XILINX_UART16550_CLOCK_HZ\t[xget_value $drv_handle "PARAMETER" "CLOCK_HZ"]"
				#dependent end
			}
			"opb_uartlite" -
			"xps_uartlite" {
				set args [xget_sw_parameter_handle $uart_handle "*"]
				foreach arg $args {
					set arg_name [xget_value $arg "NAME"]
					set arg_value [xget_value $arg "VALUE"]
					set arg_name [string map -nocase {C_ ""} $arg_name]
					#DEBUG	puts $config_file "$arg_name $arg_value"
					case $arg_name in {
						"BASEADDR" {
							puts $config_file "#define XILINX_UARTLITE_${arg_name}\t$arg_value"
							set uart_base $arg_value
						}
						"BAUDRATE" {
							puts $config_file "#define XILINX_UARTLITE_${arg_name}\t$arg_value"
						}
						"HIGHADDR" {
							set uart_end $arg_value
						}
						default	{}
					}
				}
			}
			default {
				error "Unsupported type of console - $type"
			}
		}
	}
	puts $config_file ""
	# ******************************************************************************
	# IIC driver - I suppose, only uartlite driver
	set iic [xget_sw_parameter_value $os_handle "iic"]
	if {[string match "" $iic] || [string match -nocase "none" $iic]} {
		puts $config_file "/* IIC doesn't exist */"
	} else {
		set iic_handle [xget_sw_ipinst_handle_from_processor $proc_handle $iic]
		set iic_baseaddr [xget_sw_parameter_value $iic_handle "C_BASEADDR"]
		set iic_baseaddr [format "0x%08x" $iic_baseaddr]
		set iic_freq [xget_sw_parameter_value $iic_handle "C_IIC_FREQ"]
		set iic_bit [xget_sw_parameter_value $iic_handle "C_TEN_BIT_ADR"]
		puts $config_file "/* IIC pheriphery is $iic */"
		puts $config_file "#define XILINX_IIC_0_BASEADDR\t$iic_baseaddr"
		puts $config_file "#define XILINX_IIC_0_FREQ\t$iic_freq"
		puts $config_file "#define XILINX_IIC_0_BIT\t$iic_bit"
	}
	puts $config_file ""
	# ******************************************************************************
	# GPIO configuration
	set gpio [xget_sw_parameter_value $os_handle "gpio"]
	if {[string match "" $gpio] || [string match "none" $gpio]} {
		puts $config_file "/* GPIO doesn't exist */"
	} else {
		set base_param_name [format "C_BASEADDR" $gpio]
		set gpio_handle [xget_sw_ipinst_handle_from_processor $proc_handle $gpio]
		set gpio_base [xget_sw_parameter_value $gpio_handle $base_param_name]
		set gpio_base [format "0x%08x" $gpio_base]

		set gpio_end [xget_sw_parameter_value $gpio_handle "C_HIGHADDR"]
		set gpio_end [format "0x%08x" $gpio_end]

		puts $config_file "/* GPIO is $gpio*/"
		puts $config_file "#define XILINX_GPIO_BASEADDR\t$gpio_base"
	}
	puts $config_file ""
	# ******************************************************************************
	# System memory
	set main_mem [xget_sw_parameter_value $os_handle "main_memory"]
	if {[string match "" $main_mem] || [string match "none" $main_mem]} {
		puts "ERROR main_memory not specified. Please specific main_memory"
		puts $config_file "/* Main Memory doesn't exist */"
	} else {
		set main_mem_bank [xget_sw_parameter_value $os_handle "main_memory_bank"]
		set main_mem_handle [xget_sw_ipinst_handle_from_processor $proc_handle $main_mem]
		if {[string compare -nocase $main_mem_handle ""] != 0} {
			if {[xget_hw_value $main_mem_handle] == "mpmc"} {
				set base_param_name "C_MPMC_BASEADDR"
				set high_param_name "C_MPMC_HIGHADDR"
			} else {
				set base_param_name [format "C_MEM%i_BASEADDR" $main_mem_bank]
				set high_param_name [format "C_MEM%i_HIGHADDR" $main_mem_bank]
			}
			set eram_base [xget_sw_parameter_value $main_mem_handle $base_param_name]
			set eram_end [xget_sw_parameter_value $main_mem_handle $high_param_name]
			set eram_size [expr $eram_end - $eram_base + 1]
			set eram_base [format "0x%08x" $eram_base]
			set eram_size [format "0x%08x" $eram_size]
			set eram_high [expr $eram_base + $eram_size]
			set eram_high [format "0x%08x" $eram_high]
			puts $config_file "/* Main Memory is $main_mem */"
			puts $config_file "#define XILINX_RAM_START\t$eram_base"
			puts $config_file "#define XILINX_RAM_SIZE\t$eram_size"
		 }
	}
	puts $config_file ""
	# ******************************************************************************
	# Flash memory
	set flash_mem [xget_sw_parameter_value $os_handle "flash_memory"]
	if {[string match "" $flash_mem] || [string match "none" $flash_mem]} {
		puts $config_file "/* FLASH doesn't exist $flash_mem */"
		puts "FLASH doesn't exists"
	} else {
		set flash_mem_handle [xget_sw_ipinst_handle_from_processor $proc_handle $flash_mem]
		set flash_mem_bank [xget_sw_parameter_value $os_handle "flash_memory_bank"]
		set base_param_name [format "C_MEM%i_BASEADDR" $flash_mem_bank]
		set high_param_name [format "C_MEM%i_HIGHADDR" $flash_mem_bank]
		set flash_start [xget_sw_parameter_value $flash_mem_handle $base_param_name]
		set flash_end [xget_sw_parameter_value $flash_mem_handle $high_param_name]
		set flash_size [expr $flash_end - $flash_start + 1]
		set flash_start [format "0x%08x" $flash_start]
		set flash_size [format "0x%08x" $flash_size]
		if {$eram_base < $flash_start} {
			puts $config_file "/* Flash Memory is $flash_mem */"
			puts $config_file "#define XILINX_FLASH_START\t$flash_start"
			puts $config_file "#define XILINX_FLASH_SIZE\t$flash_size"
		} else {
			error "Flash base address must be on higher address than ram memory"
		}
	}
	puts $config_file ""
	# ******************************************************************************
	# Sysace
	set sysace [xget_sw_parameter_value $os_handle "sysace"]
	if {[string match "" $sysace] || [string match "none" $sysace]} {
		puts $config_file "/* Sysace doesn't exist */"
	} else {
		puts $config_file "/* Sysace Controller is $sysace */"
		set sysace_handle [xget_sw_ipinst_handle_from_processor $proc_handle $sysace]
		set args [xget_sw_parameter_handle $sysace_handle "*"]
		foreach arg $args {
			set arg_name [xget_value $arg "NAME"]
			set arg_name [string map -nocase {C_ ""} $arg_name]
			set arg_value [xget_value $arg "VALUE"]
			#	puts $config_file "$arg_name $arg_value"
			switch $arg_name {
				"BASEADDR" {
					#	set arg_value [xget_value $arg "VALUE"]
					puts $config_file "#define XILINX_SYSACE_${arg_name}\t$arg_value"
					set sysace_base $arg_value
				}
				"MEM_WIDTH" {
					#	set arg_value [xget_value $arg "VALUE"]
					puts $config_file "#define XILINX_SYSACE_${arg_name}\t$arg_value"
				}
				"HIGHADDR" {
					#	set arg_value [xget_value $arg "VALUE"]
					set sysace_end $arg_value
				}
				"HW_VER" {
					#	puts "revize $arg_value"
				}
				default {}
			}
		}
	}
	puts $config_file ""
	# ******************************************************************************
	# Ethernet
	set ethernet [xget_sw_parameter_value $os_handle "ethernet"]
	if {[string match "" $iic] || [string match -nocase "none" $ethernet]} {
		puts $config_file "/* Ethernet doesn't exist */"
	} else {
		set ethernet_handle [xget_sw_ipinst_handle_from_processor $proc_handle $ethernet]
		set ethernet_name [xget_value $ethernet_handle "VALUE"]
		#	puts "DEBUG $ethernet_name"
		puts $config_file "/* Ethernet controller is $ethernet */"

		switch $ethernet_name {
			"opb_ethernet" {
				set args [xget_sw_parameter_handle $ethernet_handle "*"]
				foreach arg $args {
					set arg_name [xget_value $arg "NAME"]
					set arg_value [xget_value $arg "VALUE"]
					set arg_name [string map -nocase {C_ ""} $arg_name]
					#DEBUG	puts $config_file "$arg_name $arg_value"
					case $arg_name in {
						{"MII_EXIST" "DMA_PRESENT" "HALF_DUPLEX_EXIST"} {
							puts $config_file "#define XILINX_EMAC_${arg_name}\t$arg_value"
						}
						"BASEADDR" {
							puts $config_file "#define XILINX_EMAC_${arg_name}\t$arg_value"
							set ethernet_base $arg_value
						}
						"HIGHADDR" {
							set ethernet_end $arg_value
						}
						default {}
					}
				 }
			}
			"xps_ll_temac" {
				set args [xget_sw_parameter_handle $ethernet_handle "*"]
				foreach arg $args {
					set arg_name [xget_value $arg "NAME"]
					set arg_value [xget_value $arg "VALUE"]
					set arg_name [string map -nocase {C_ ""} $arg_name]
					#DEBUG	puts $config_file "$arg_name $arg_value"
					case $arg_name in {
						{"MII_EXIST" "DMA_PRESENT" "HALF_DUPLEX_EXIST"} {
							puts $config_file "#define XILINX_LLEMAC_${arg_name}\t$arg_value"
						}
						"BASEADDR" {
							puts $config_file "#define XILINX_LLEMAC_${arg_name}\t$arg_value"
							set ethernet_base $arg_value
						}
						"HIGHADDR" {
							set ethernet_end $arg_value
						}
						default {}
					}
				}
			}
			"opb_ethernetlite" {
				set args [xget_sw_parameter_handle $ethernet_handle "*"]
				foreach arg $args {
					set arg_name [xget_value $arg "NAME"]
					set arg_value [xget_value $arg "VALUE"]
					set arg_name [string map -nocase {C_ ""} $arg_name]
					#DEBUG	puts $config_file "$arg_name $arg_value"
					case $arg_name in {
						{"BASEADDR" "TX_PING_PONG" "RX_PING_PONG"} {
						puts $config_file "#define XILINX_EMACLITE_${arg_name}\t$arg_value"
						}
						efault {}
					}
				}
			}
			default {
				error "Unsupported ethernet periphery - $ethernet_name"
			}
		}
	}
	#*******************************************************************************
	# U-BOOT position in memory
	set text_base [xget_sw_parameter_value $os_handle "uboot_position"]
	set text_base [format "0x%08x" $text_base]
	puts $config_file2 ""
	if {$text_base == 0} {
		if {[llength $eram_base] != 0 } {
			puts $config_file2 "TEXT_BASE = $eram_base"
			puts "WARNING automatic U-BOOT position"
		} else {
			error "Main memory is not defined"
		}
	} elseif {![catch {unset $eram_base}]} {
		if {$eram_base < $text_base && $eram_high > $text_base} {
			#			puts $config_file2 "# TEXT BASE "
			puts $config_file2 "TEXT_BASE = $text_base"
			puts $config_file2 ""
			# print system clock
			#	set sw [xget_sw_parameter_value $proc_handle "HW_INSTANCE"]
			# FIXME Parameters for Microblaze from MHS files
			# look at Microblaze SW manual
			# print microblaze params
			#	set hwproc_handle [xget_handle $proc_handle "IPINST"]
			#	set args [xget_hw_parameter_handle $hwproc_handle "*"]
			#	set proctype [xget_value $hwproc_handle "OPTION" "IPNAME"]
		} else {
			error "ERROR u-boot position is out of range $eram_base - $eram_high"
		}
	}
}

proc headerm {ufile} {
	puts $ufile "\#"
	puts $ufile "\# (C) Copyright 2007 Michal Simek"
	puts $ufile "\#"
	puts $ufile "\# Michal SIMEK <monstr@monstr.eu>"
	puts $ufile "\#"
	puts $ufile "\# This program is free software; you can redistribute it and/or"
	puts $ufile "\# modify it under the terms of the GNU General Public License as"
	puts $ufile "\# published by the Free Software Foundation; either version 2 of"
	puts $ufile "\# the License, or (at your option) any later version."
	puts $ufile "\#"
	puts $ufile "\# This program is distributed in the hope that it will be useful,"
	puts $ufile "\# but WITHOUT ANY WARRANTY; without even the implied warranty of"
	puts $ufile "\# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the"
	puts $ufile "\# GNU General Public License for more details."
	puts $ufile "\#"
	puts $ufile "\# You should have received a copy of the GNU General Public License"
	puts $ufile "\# along with this program; if not, write to the Free Software"
	puts $ufile "\# Foundation, Inc., 59 Temple Place, Suite 330, Boston,"
	puts $ufile "\# MA 02111-1307 USA"
	puts $ufile "\#"
	puts $ufile "\# CAUTION: This file is automatically generated by libgen."
	puts $ufile "\# Version: [xget_swverandbld]"
	puts $ufile "\#"
	puts $ufile ""
}

proc headerc {ufile} {
	puts $ufile "/*"
	puts $ufile " * (C) Copyright 2007 Michal Simek"
	puts $ufile " *"
	puts $ufile " * Michal SIMEK <monstr@monstr.eu>"
	puts $ufile " *"
	puts $ufile " * This program is free software; you can redistribute it and/or"
	puts $ufile " * modify it under the terms of the GNU General Public License as"
	puts $ufile " * published by the Free Software Foundation; either version 2 of"
	puts $ufile " * the License, or (at your option) any later version."
	puts $ufile " *"
	puts $ufile " * This program is distributed in the hope that it will be useful,"
	puts $ufile " * but WITHOUT ANY WARRANTY; without even the implied warranty of"
	puts $ufile " * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the"
	puts $ufile " * GNU General Public License for more details."
	puts $ufile " *"
	puts $ufile " * You should have received a copy of the GNU General Public License"
	puts $ufile " * along with this program; if not, write to the Free Software"
	puts $ufile " * Foundation, Inc., 59 Temple Place, Suite 330, Boston,"
	puts $ufile " * MA 02111-1307 USA"
	puts $ufile " *"
	puts $ufile " * CAUTION: This file is automatically generated by libgen."
	puts $ufile " * Version: [xget_swverandbld]"
	puts $ufile " */"
	puts $ufile ""
}

# from xillib_hw.tcl
#
# Return the handle to the bus interface associated with this address parameter.
# Parameters: periph_handle: peripheral handle
#	addr_name: name of an address parameter
#
proc xget_addrparam_busif {periph_handle addr_param_name} {
	set baseaddr_handle [xget_handle $periph_handle "PARAMETER" $addr_param_name]
	# test on null parameter
	if { [llength $baseaddr_handle] == 0 } {
		return ""
	}

	set baseaddr_bushandle [xget_handle $baseaddr_handle "SUBPROPERTY" "BUS"]
	puts "DEBUG $baseaddr_bushandle =="
	if { [llength $baseaddr_bushandle] == 0 } {
		# See if this address has a pair
		if {[string match *_BASEADDR $addr_param_name]} {
			set addr_pair [string range $addr_param_name 0 [expr [string last "_BASEADDR" $addr_param_name] -1]]
			append addr_pair "_HIGHADDR"
		} elseif {[string match *_HIGHADDR $addr_param_name]} {
			set addr_pair [string range $addr_param_name 0 [expr [string last "_HIGHADDR" $addr_param_name] -1]]
			append addr_pair "_BASEADDR"
		} else {
			return ""
		}
		set baseaddr_handle [xget_handle $periph_handle "PARAMETER" $addr_pair]
		set baseaddr_bushandle [xget_handle $baseaddr_handle "SUBPROPERTY" "BUS"]
		if { [llength $baseaddr_bushandle] == 0 } {
			return ""
		}
	}
	set baseaddr_busname [xget_value $baseaddr_handle "SUBPROPERTY" "BUS"]
	#	set baseaddr_busname [xget_value $baseaddr_bushandle "VALUE"]
	set busif_handle [xget_handle $periph_handle "BUS_INTERFACE" $baseaddr_busname]
	return $busif_handle
}

#test for peripheral - if is correct setting system bus
proc test_buses {system_bus handle bus_type} {
	set bus [xget_handle $handle "BUS_INTERFACE" $bus_type]
	if { [llength $bus] == 0 } {
		return 1
	}
	set bus [xget_value $bus "VALUE"]
	if { $bus != $system_bus} {
		error "Periphery $handle is connected to another system bus $bus ----"
		return 0
	} else {
		set name [xget_value $handle "NAME"]
		puts "$name has correct system_bus $system_bus"
	}
	return 1
}

proc xdefine_include_file2 {drv_handle file_name drv_string args} {
	foreach arg $args {
		set value [xget_value $drv_handle "PARAMETER" $arg]
		if {[llength $value] == 0} {
			error "Zero size of PARAMETER. Please append argument args."
		} else {
			#	puts $file_name "#defineoo [xget_dname $drv_string $arg] [xget_value $drv_handle "PARAMETER" $arg]"
			return [xget_value $drv_handle "PARAMETER" $arg]
		}
	}
}

proc get_intc_signals {intc} {
	set signals [split [xget_hw_port_value $intc "intr"] "&"]
	set intc_signals {}
	foreach signal $signals {
		lappend intc_signals [string trim $signal]
	}
	return $intc_signals
}

# Get interrupt number
proc get_intr {per_handle intc intc_value port_name} {
	if {![string match "" $intc] && ![string match -nocase "none" $intc]} {
		set intc_signals [get_intc_signals $intc]
		set port_handle [xget_hw_port_handle $per_handle "$port_name"]
		set interrupt_signal [xget_value $port_handle "VALUE"]
		set index [lsearch $intc_signals $interrupt_signal]
		if {$index == -1} {
			return -1
		} else {
			# interrupt 0 is last in list.
			return [expr [llength $intc_signals] - $index - 1]
		}
	} else {
		return -1
	}
}

proc slaveip {slave intc devicetype params {baseaddr_prefix ""} } {
	set name [xget_hw_name $slave]
	set type [xget_hw_value $slave]
	if {$devicetype == ""} {
		set devicetype $type
	}

	set baseaddr [scan_int_parameter_value $slave [format "C_%sBASEADDR" $baseaddr_prefix]]
	set highaddr [scan_int_parameter_value $slave [format "C_%sHIGHADDR" $baseaddr_prefix]]
	set hw_ver [xget_hw_parameter_value $slave "HW_VER"]

	set ip_node {}
	set ip_node [gen_compatible_property $ip_node $name $type $hw_ver]
	lappend ip_node [list "xlnx,edk-name" string "$name"]
	set ip_node [gen_reg_property $ip_node $name $baseaddr $highaddr]

	if {$slave == $intc} {
		lappend ip_node [list "linux,phandle" hexint 257]
	}

	# Generate the parameters
	set ip_node [gen_params $ip_node $slave $params]

	return [list [format_ip_name $devicetype $baseaddr] tree $ip_node]
}

proc gen_intc {slave intc devicetype param} {
	set tree [slaveip $slave $intc $devicetype $param]
	set intc_name [lindex $tree 0]
	set intc_node [lindex $tree 2]

	# Tack on the interrupt-specific tags.
	lappend intc_node [list \#interrupt-cells hexint 2]
	lappend intc_node [list interrupt-controller empty empty]
	return [list $intc_name tree $intc_node]
}

proc slaveip_intr {slave intc interrupt_port_list devicetype params {baseaddr_prefix ""} } {
	set pocet [scan_int_parameter_value $intc "C_NUM_INTR_INPUTS"]
	set pocet [expr $pocet - 1]
	set opb_intc [xget_hw_name $intc]
	#	puts "---$slave $opb_intc $pocet $intr [llength $interrupt_port_list]---"
	set tree [slaveip $slave $intc $devicetype $params $baseaddr_prefix]
	set ip_name [lindex $tree 0]
	set ip_node [lindex $tree 2]

	set interrupt_list {}
	foreach in $interrupt_port_list {
		set irq [get_intr $slave $intc $pocet $in]
		if {![string match $irq "-1"]} {
			lappend interrupt_list "$irq" 0
		}
	}
	if {[llength $interrupt_list] != 0} {
		lappend ip_node [list "interrupts" hexinttuple $interrupt_list ]
		lappend ip_node [list "interrupt-parent" hexint 257]
	}
	return [list $ip_name tree $ip_node]

}

#
#get handle to interrupt controller from CPU handle
#
proc get_handle_to_intc {proc_handle port_name} {
	#one CPU handle
	set hwproc_handle [xget_handle $proc_handle "IPINST"]
	#hangle to mhs file
	set mhs_handle [xget_hw_parent_handle $hwproc_handle]
	#get handle to interrupt port on Microblaze
	set intr_port [xget_value $hwproc_handle "PORT" $port_name]
	if { [llength $intr_port] == 0 } {
		error "CPU has not connection to Interrupt controller"
	}
	#	set sink_port [xget_hw_connected_ports_handle $mhs_handle $intr_port "sink"]
	#	set sink_name [xget_hw_name $sink_port]
	#get source port periphery handle - on interrupt controller
	set source_port [xget_hw_connected_ports_handle $mhs_handle $intr_port "source"]
	#get interrupt controller handle
	set intc [xget_hw_parent_handle $source_port]
	#	set name [xget_hw_name $intc]
	#	puts "$intc $name"
	return $intc
}

#retun number of tabulator
proc tt {number} {
	set tab ""
	for {set x 0} {$x < $number} {incr x} {
		set tab "$tab\t"
		#	puts ".$tab..."
	}
	return $tab
}

proc gener_slave {tree slave intc} {
	set name [xget_hw_name $slave]
	set type [xget_hw_value $slave]
	switch -exact $type {
		"opb_intc" -
		"xps_intc" {
			# Interrupt controllers
			lappend tree [gen_intc $slave $intc "interrupt-controller" "C_NUM_INTR_INPUTS"]
		}
		"mdm" -
		"opb_mdm" {
			# Microblaze debug
			lappend tree [slaveip_intr $slave $intc [interrupt_list $slave] "" [default_parameters $slave] ]
			#"C_MB_DBG_PORTS C_UART_WIDTH C_USE_UART"]
		}
		"xps_uartlite" -
		"opb_uartlite" {
			lappend tree [slaveip_intr $slave $intc [interrupt_list $slave] "serial" [default_parameters $slave] ]
			#"BAUDRATE DATA_BITS CLK_FREQ ODD_PARITY USE_PARITY"]
		}
		"xps_uart16550" -
		"plb_uart16550" -
		"opb_uart16550" {
			lappend tree [slaveip_intr $slave $intc [interrupt_list $slave] "serial" [default_parameters $slave] ]
			#"BAUDRATE DATA_BITS CLK_FREQ ODD_PARITY USE_PARITY"]
		}
		"xps_timer" -
		"opb_timer" {
			lappend tree [slaveip_intr $slave $intc [interrupt_list $slave] "" [default_parameters $slave] ]
			#"C_COUNT_WIDTH C_ONE_TIMER_ONLY"]
		}
		"xps_sysace" -
		"opb_sysace" {
			lappend tree [slaveip_intr $slave $intc [interrupt_list $slave] "" [default_parameters $slave] ]
			#"MEM_WIDTH"]
		}
		"opb_ethernet" -
		"plb_ethernet" -
		"opb_ethernetlite" -
		"xps_ethernetlite" -
		"plb_temac" {
			# 'network' type
			lappend tree [slaveip_intr $slave $intc [interrupt_list $slave] "network" [default_parameters $slave]]
		}
		"xps_ll_temac" {
			# We need to handle this specially, to notify the driver
			# about the connected LL connection.
			lappend tree [slaveip_intr $slave $intc "TemacIntc0_Irpt TemacIntc1_Irpt" "network" [default_parameters $slave]]
		}
		"opb_ps2_dual_ref" {
			# We should handle this specially, to report the two
			# independent ports.
			lappend tree [slaveip_intr $slave $intc "Sys_Intr1 Sys_Intr2" "" [default_parameters $slave] ""]
		}
		"opb_ac97_controller_ref" {
			# We should handle this specially, to report the two
			# interrupts in the right order.
			lappend tree [slaveip_intr $slave $intc "Playback_Interrupt Record_Interrupt" "" ""]
		}
		"plb_bram_if_cntlr" -
		"opb_bram_if_cntlr" -
		"opb_cypress_usb" -
		"plb_ddr" -
		"plb_ddr2" -
		"plb_emc" -
		"opb_sdram" -
		"opb_ddr" -
		"opb_emc" -
		"mch_opb_ddr" -
		"mch_opb_ddr2" -
		"mch_opb_emc" {
			# Do nothing.. this is handled by the 'memory' special case.
		}
		"mpmc" {
			# We should handle this specially, to report the DMA
			# ports. This is a hack that happens to work for the
			# design I have. Note that we don't use the default
			# parameters here because of the slew of parameters the
			# mpmc has.
			# lappend tree [slaveip_intr $slave $intc "SDMA3_Rx_IntOut SDMA3_Tx_IntOut" "" "" "SDMA_CTRL_"]
		}
		"opb2plb_bridge" {
			# Hmm.. how do we represent this?
			#	set tree [bus_bridge $tree $slave $intc "MPLB" "C_RNG"]
		}
		"plb2opb_bridge" {
			set tree [bus_bridge $tree $slave $intc "MOPB" "C_RNG"]
		}
		"opb_opb_lite" {
			set tree [bus_bridge $tree $slave $intc "MOPB" "C_DEC"]
		}
		"microblaze" {
			puts "Other Microblaze CPU $name=$type"
			lappend tree [gen_microblaze $slave [default_parameters $slave]]
		}
		"ppc405" {
			puts "Other PowerPC405 CPU $name=$type"
			lappend tree [gen_ppc405 $slave [default_parameters $slave]]
		}
		default {
			# *Most* IP should be handled by this default case.
			if {[catch {lappend tree [slaveip_intr $slave $intc [interrupt_list $slave] "" [default_parameters $slave] "" ]} {error}]} {
			puts "Default handling for unknown IP $name ($type) Failed... It won't show up in the device tree."
			}
		}
	}
	return $tree
}

proc memory {slave baseaddr_prefix params} {
	set name [xget_hw_name $slave]
	set type [xget_hw_value $slave]
	set par [xget_hw_parameter_handle $slave "*"]
	set hw_ver [xget_hw_parameter_value $slave "HW_VER"]

	set ip_node {}

	set baseaddr [scan_int_parameter_value $slave [format "C_%sBASEADDR" $baseaddr_prefix]]
	set highaddr [scan_int_parameter_value $slave [format "C_%sHIGHADDR" $baseaddr_prefix]]

	set ip_node [gen_reg_property $ip_node $name $baseaddr $highaddr]
	set ip_node [gen_compatible_property $ip_node $name $type $hw_ver]
	lappend ip_node [list "xlnx,edk-name" string "$name"]
	lappend ip_node [list "device_type" string "memory"]
	set ip_node [gen_params $ip_node $slave $params]
	return [list [format_ip_name memory $baseaddr] tree $ip_node]
}

proc test {name args} {
	puts "$name [llength $args]"

	if {[llength $name] > 1} {
		foreach in $args {
			puts "$in"
		}
	}
}

proc gen_ppc405 {tree hwproc_handle params} {
	set out ""
	variable cpunumber

	set cpu_type [xget_hw_value $hwproc_handle]
	set hw_ver [xget_hw_parameter_value $hwproc_handle "HW_VER"]

	set cpus_node {}
	set proc_node {}
	lappend proc_node [list device_type string cpu]
	# Fixme: get the right clock frequency.
	lappend proc_node [list clock-frequency int 300000000]
	lappend proc_node [list timebase-frequency int 100000000]
	lappend proc_node [list reg int $cpunumber]
	lappend proc_node [list i-cache-size hexint [expr 0x4000]]
	lappend proc_node [list i-cache-line-size hexint 32]
	lappend proc_node [list d-cache-size hexint [expr 0x4000]]
	lappend proc_node [list d-cache-line-size hexint 32]
	set proc_node [gen_params $proc_node $hwproc_handle $params]

	lappend cpus_node [list "PowerPC,405@$cpunumber" "tree" "$proc_node"]
	lappend cpus_node [list \#size-cells hexint 0]
	lappend cpus_node [list \#address-cells hexint 1]
	set cpunumber [expr $cpunumber + 1]
	lappend cpus_node [list \#cpus hexint "$cpunumber" ]
	lappend tree [list cpus tree "$cpus_node"]
	return $tree
}

proc gen_microblaze {tree hwproc_handle params} {
	set out ""
	variable cpunumber

	set cpu_type [xget_hw_value $hwproc_handle]

	set icache_size [scan_int_parameter_value $hwproc_handle "C_CACHE_BYTE_SIZE"]
	set icache_base [scan_int_parameter_value $hwproc_handle "C_ICACHE_BASEADDR"]
	set icache_high [scan_int_parameter_value $hwproc_handle "C_ICACHE_HIGHADDR"]
	set dcache_size [scan_int_parameter_value $hwproc_handle "C_DCACHE_BYTE_SIZE"]
	set dcache_base [scan_int_parameter_value $hwproc_handle "C_DCACHE_BASEADDR"]
	set dcache_high [scan_int_parameter_value $hwproc_handle "C_DCACHE_HIGHADDR"]
	set hw_ver [xget_hw_parameter_value $hwproc_handle "HW_VER"]

	set cpus_node {}
	set proc_node {}
	lappend proc_node [list device_type string cpu]
	# Fixme: get the right clock frequency.
	lappend proc_node [list clock-frequency int 100000000]
	lappend proc_node [list timebase-frequency int 100000000]
	lappend proc_node [list reg int 0]
	if { [llength $icache_size] != 0 } {
		lappend proc_node [list i-cache-baseaddr hexint $icache_base]
		lappend proc_node [list i-cache-highaddr hexint $icache_high]
		lappend proc_node [list i-cache-size hexint $icache_size]
		lappend proc_node [list i-cache-line-size hexint 16]
	}
	if { [llength $dcache_size] != 0 } {
		lappend proc_node [list d-cache-baseaddr hexint $dcache_base]
		lappend proc_node [list d-cache-highaddr hexint $dcache_high]
		lappend proc_node [list d-cache-size hexint $dcache_size]
		lappend proc_node [list d-cache-line-size hexint 16]
	}

	#-----------------------------
	# generating additional parameters
	# the list of Microblaze parameters
	set proc_node [gen_params $proc_node $hwproc_handle $params]

	#-----------------------------
	lappend cpus_node [list [format_ip_name "$cpu_type,$hw_ver" $cpunumber] "tree" "$proc_node"]
	lappend cpus_node [list \#size-cells hexint 0]
	lappend cpus_node [list \#address-cells hexint 1]
	set cpunumber [expr $cpunumber + 1]
	lappend cpus_node [list \#cpus hexint "$cpunumber" ]
	lappend tree [list cpus tree "$cpus_node"]
	return $tree
}

proc gen_memories {tree hwproc_handle} {
	set mhs_handle [xget_hw_parent_handle $hwproc_handle]
	set ip_handles [xget_hw_ipinst_handle $mhs_handle "*"]
	foreach slave $ip_handles {
		set name [xget_hw_name $slave]
		set type [xget_hw_value $slave]
		switch $type {
			"plb_bram_if_cntlr" -
			"opb_bram_if_cntlr" -
			"opb_sdram" {
				# Handle bankless memories.
				lappend tree [memory $slave "" ""]
			}
			"opb_cypress_usb" -
			"plb_ddr" -
			"plb_ddr2" -
			"plb_emc" -
			"opb_sdram" -
			"opb_ddr" -
			"opb_emc" -
			"mch_opb_ddr" -
			"mch_opb_ddr2" -
			"mch_opb_emc" {
				# Handle memories with 'banks'. Generate one memory
				# node for each bank.
				set count [scan_int_parameter_value $slave "C_NUM_BANKS_MEM"]
				if { [llength $count] == 0 } {
					set count 1
				}
				for {set x 0} {$x < $count} {incr x} {
					lappend tree [memory $slave [format "MEM%d_" $x] ""]
				}
			}
			"mpmc" {
				lappend tree [memory $slave "MPMC_" ""]
			}
		}
	}
	return $tree
}

# Populates a bus node with components connected to the given slave
# and adds it to the given tree
#
# tree		: Tree to populate

# slave_handle	: The slave to use as a starting point, this is
# typically the root processor or a previously traversed bus bridge.

# intc_handle	: The interrupt controller associated with the
# processor. Slave will have an interrupts node relative to this
# controller.

# face		: The name of the port of the slave that is connected to the bus.

# baseaddr_prefix : prefix used to query the slave for address
# translation ranges, or empty to generate a 1:1 translation.
proc bus_bridge {tree slave intc_handle face baseaddr_prefix} {
	puts "+++++++++++ $slave ++++++++"
	set busif_handle [xget_hw_busif_handle $slave $face]
	if {[llength $busif_handle] == 0} {
		return $tree
	}
	set bus_name [xget_hw_value $busif_handle]

	#	puts "--$busif_handle"
	set mhs_handle [xget_hw_parent_handle $slave]
	set bus_handle [xget_hw_ipinst_handle $mhs_handle $bus_name]

	if {[llength $bus_handle] == 0} {
		return $tree
	}
	set bus_type [xget_hw_value $bus_handle]
	set hw_ver [xget_hw_parameter_value $bus_handle "HW_VER"]

	set master_ifs [xget_hw_connected_busifs_handle $mhs_handle $bus_name "master"]
	set slave_ifs [xget_hw_connected_busifs_handle $mhs_handle $bus_name "slave"]
	set bus_ip_handles {}
	# Compose peripherals & cleaning
	foreach if $master_ifs {
		set ip_handle [xget_hw_parent_handle $if]
		puts "-master $ip_handle [xget_hw_name $if] [xget_hw_value $if] [xget_hw_name $ip_handle]"

		# If its not already in the list, and its not the bridge, then
		# append it.
		if {$ip_handle != $slave} {
			if {[lsearch $bus_ip_handles $ip_handle] == -1} {
				lappend bus_ip_handles $ip_handle
			}
		}
	}

	foreach if $slave_ifs {
		set ip_handle [xget_hw_parent_handle $if]
		puts "-slave $ip_handle [xget_hw_name $if] [xget_hw_value $if] [xget_hw_name $ip_handle]"

		# If its not already in the list, and its not the bridge, then
		# append it.
		if {$ip_handle != $slave} {
			if {[lsearch $bus_ip_handles $ip_handle] == -1} {
				 lappend bus_ip_handles $ip_handle
			}
		}
	}
	variable periphery_array

	# Start generating the node for the bus.
	set bus_node {}

	# Generate the ranges
# 	if {$baseaddr_prefix != ""} {
# 		set count [scan_int_parameter_value $slave "C_NUM_ADDR_RNG"]
# 		if { [llength $count] == 0 } {
# 			set count 1
# 		}
# 		for {set x 0} {$x < $count} {incr x} {
# 			set baseaddr [scan_int_parameter_value $slave [format "%s%s_BASEADDR" $baseaddr_prefix $x]]
# 			set highaddr [scan_int_parameter_value $slave [format "%s%s_HIGHADDR" $baseaddr_prefix $x]]

# 			set bus_node [gen_ranges_property $bus_node $bus_name $x $baseaddr $highaddr]
# 		}
# 	 } else {
# 		lappend bus_node [list ranges empty empty]
# 	 }

	# Populate with all the slaves.
	foreach ip $bus_ip_handles {
		# If we haven't already generated this ip
		if {[lsearch $periphery_array $ip] == -1} {
			set bus_node [gener_slave $bus_node $ip $intc_handle]
			lappend periphery_array $ip
		}
	}

	lappend bus_node [list \#size-cells hexint 1]
	lappend bus_node [list \#address-cells hexint 1]
	set bus_node [gen_compatible_property $bus_node $bus_name $bus_type $hw_ver]

	lappend tree [list $bus_type tree $bus_node]
	return $tree
}

# Return a sorted list of all the port names that we think are
# interrupts (i.e. those tagged in the mpd with SIGIS=INTERRUPT)
proc interrupt_list {ip_handle} {
	set port_handles [xget_hw_port_handle $ip_handle "*"]
	set interrupt_ports {}
	foreach port $port_handles {
		set name [xget_value $port "NAME"]
		set list [xget_hw_subproperty_handle $port "SIGIS"]
		set sigis [xget_hw_subproperty_value $port "SIGIS"]
		if {[string match $sigis "INTERRUPT"]} {
			lappend interrupt_ports $name
		}
	}
	return [lsort $interrupt_ports]
}

# Return a list of all the parameter names for the given ip that
# should be reported in the device tree for generic IP. This list
# includes all the parameter names, except those that are handled
# specially, such as the instance name, baseaddr, etc.
proc default_parameters {ip_handle} {
	set par_handles [xget_hw_parameter_handle $ip_handle "*"]
	set params {}
	foreach par $par_handles {
		set par_name [xget_hw_name $par]
		# Ignore some parameters that are always handled specially
		switch -glob $par_name {
			"INSTANCE" -
			"C_FAMILY" -
			"*BASEADDR" -
			"*HIGHADDR" -
			"C_SPLB*" -
			"HW_VER" {}
			default { lappend params $par_name }
		}
	}
	return $params
}

proc scan_int_parameter_value {ip_handle name} {
	set param_handle [xget_hw_parameter_handle $ip_handle $name]
	if {$param_handle == ""} {
		error "Can't find parameter $name in [xget_hw_name $ip_handle]"
		return 0
	}
	return [expr [xget_hw_value $param_handle]]
}

proc format_xilinx_name {name} {
	return "xlnx,[format_name $name]"
}
proc format_param_name {name} {
	# This fails if the parameter name contains an internal C_
	if {[string match [string range $name 0 1] "C_"]} {
		set name [string range $name 2 [string length $name]]
	}
	return [format_xilinx_name $name]
}

proc format_ip_name {devicetype baseaddr} {
	return "[format_name $devicetype@[format %x $baseaddr]]"
}

proc gen_all_params {tree handle} {
	set par_handles [xget_hw_parameter_handle $handle "*"]
	foreach par_handle $par_handles {
		set par_name [xget_hw_name $par_handle]
		set par_value [xget_hw_value $par_handle]
		lappend tree [list [format_param_name $par_name] hexint $par_value]
	}
	return $tree
}

proc gen_params {tree handle params} {
	foreach i $params {
		set par_handle [xget_hw_parameter_handle $handle "$i"]
		if {$par_handle == ""} {
			puts "Unknown parameter name $i"
		} else {
			set par_name [xget_hw_name $par_handle]
			set par_value [xget_hw_value $par_handle]
			lappend tree [list [format_param_name $par_name] hexint $par_value]
		}
	}
	return $tree
}

proc gen_compatible_property {node nodename type hw_ver} {
	set namewithver [format "%s_%s" $type $hw_ver]
	lappend node [list "compatible" stringtuple [list [format_xilinx_name "$namewithver"] [format_xilinx_name "$type"]]]
	return $node
}

proc gen_ranges_property {node nodename index baseaddr highaddr} {
	if { ![llength $baseaddr] || ![llength $highaddr] } {
		error "Bad address range $nodename"
	}
	if {[string match $highaddr "0x00000000"]} {
		return $node
	}
	set size [expr $highaddr - $baseaddr + 1]
	if { $size < 0 } {
		error "Bad highaddr for $nodename"
	}
	lappend node [list "ranges" hexinttuple [list $index $baseaddr $size] ]
	return $node
}

proc gen_reg_property {node nodename baseaddr highaddr} {
	if { ![llength $baseaddr] || ![llength $highaddr] } {
		error "Bad address range $nodename"
	}
	if {[string match $highaddr "0x00000000"]} {
		return $node
	}
	set size [expr $highaddr - $baseaddr + 1]
	if { $size < 0 } {
		error "Bad highaddr for $nodename"
	}
	lappend node [list "reg" hexinttuple [list $baseaddr $size] ]
	return $node
}

proc format_name {par_name} {
	set par_name [string tolower $par_name]
	set par_name [string map -nocase {"_" "-"} $par_name]
	return $par_name
}

proc write_value {file indent type value} {
	if {[catch {
		if {$type == "int"} {
			#	puts -nonewline $file "= [format %d $value]"
			puts -nonewline $file "= <[format %x $value]>"
		} elseif {$type == "hexint"} {
			puts -nonewline $file "= <[format %x $value]>"
		} elseif {$type == "empty"} {
		} elseif {$type == "hexinttuple"} {
			puts -nonewline $file "= < "
			foreach element $value {
				 puts -nonewline $file "[format %x $element] "
			}
			puts -nonewline $file ">"
		} elseif {$type == "string"} {
			puts -nonewline $file "= \"$value\""
		} elseif {$type == "stringtuple"} {
			puts -nonewline $file "= "
			set first true
			foreach element $value {
				 if {$first != true} { puts -nonewline $file ", " }
				 puts -nonewline $file "\"$element\""
				 set first false
			}
		} elseif {$type == "tree"} {
			write_tree $indent $file $value
		} else {
			puts "unknown type $type"
		}
	}]} {
		puts -nonewline $file "= \"$value\""
	}
	puts $file ";"
}

proc write_nodes {indent file tree} {
	set tree [lsort -index 0 $tree]
	foreach node $tree {
		if { [llength $node] == 3} {
				set name [lindex $node 0]
				set type [lindex $node 1]
				set value [lindex $node 2]
				puts -nonewline $file "[tt [expr $indent + 1]]$name "
				write_value $file [expr $indent + 1] $type $value
		} else {
				puts "Error_bad_tree_node length = [llength $node], $node"
		}
	}
}

proc write_tree {indent file tree} {
	puts $file "{"
	set trees {}
	set nontrees {}
	foreach node $tree {
		if { [string match [lindex $node 1] "tree"]} {
			lappend trees $node
		} else {
			lappend nontrees $node
		}
	}
	write_nodes $indent $file $nontrees
	write_nodes $indent $file $trees

	puts -nonewline $file "[tt $indent]} "
}
